/* ddsconvert - Convert DXT1/3/5 textures in DirectDraw Surface (DDS) files to
 *              uncompressed TGA
 * Author: Alex Stewart <alex@foogod.com>
 * Version: 1.0
 * Date: Dec 3, 2004
 * 
 * This code has been placed into the public domain by the author.
 *
 * Description:
 *
 * This program was created mainly for working with UnrealEd, the editor
 * program for the Unreal series of game engines.  UnrealEd can use DXT
 * compressed textures internally, and supports exporting DXTC textures to DDS
 * (DirectDraw Surface) files.  Unfortunately, most other graphics applications
 * do not understand DDS files or DXT compression.  This utility therefore
 * converts DXT1, DXT3, and DXT5 compressed images (the only forms of DXTC
 * supported by UnrealEd) in DDS files to a more widely accepted format (TGA).
 * As TGA is also a supported import format in UnrealEd, this also makes for an
 * easy way to un-compress DXT textures back to RGBA8 (which UnrealEd does not
 * support directly).
 *
 * Note that this program should be sufficient for most UnrealEd-related uses,
 * but is still fairly simplistic.  It does not handle DXT2/4 formats, it does
 * not attempt to extract mipmaps, cubemaps, etc, from DDS files, will only
 * convert one-way from DDS to TGA, and so on.  Those looking for more
 * sophisticated tools might wish to look into Nvidia's texture tools, or the
 * Microsoft DirectX SDK, both of which have free utilities for manipulating
 * DDS/DXT information in more flexible ways.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>

#define WINDOW_TITLE "DXTC Conversion"

#ifndef WIN32_GUI

/* We're compiling the normal command-line version.  Have messages go to
 * stdout/stderr */

#define infomsg(fmt, args...) { printf(fmt "\n", ## args); fflush(stdout); }
#define usermsg(fmt, args...) { printf(fmt "\n", ## args); fflush(stdout); }
#define errmsg(fmt, args...) fprintf(stderr, "Error: " fmt "\n", ## args)
#define oserrmsg(str) perror(str)

#else

/* We're compiling this for the "_w" (Windows GUI mode) executable.  Do user
 * interaction via dialog boxes and such */

#include <windows.h>

#define infomsg(fmt, args...) /* Don't show info messages in GUI mode */
#define usermsg(fmt, args...) dialog_box(MB_OK | MB_ICONINFORMATION, fmt, ## args)
#define errmsg(fmt, args...) dialog_box(MB_OK | MB_ICONERROR, fmt, ## args)
#define oserrmsg(str) dialog_box(MB_OK | MB_ICONERROR, "%s: %s", str, strerror(errno))

const char window_title[] = WINDOW_TITLE;

#define MSGBOX_BUFSIZE 1024

int dialog_box(int flags, const char *fmt, ...) {
  va_list ap;
  char buf[MSGBOX_BUFSIZE];

  va_start(ap, fmt);
  vsnprintf(buf, MSGBOX_BUFSIZE, fmt, ap);
  return MessageBox(NULL, buf, window_title, flags);
}

/* This is for the GetSaveFileName call in main().  For some reason, it doesn't
 * like it if this structure is on the stack, so we'll put it here instead.  I
 * still don't really understand it, but that's Windows for ya.. */

OPENFILENAME ofn;

#endif

typedef unsigned int uint32;
typedef unsigned short uint16;
typedef unsigned char uint8;


/* ====================== DDS File Format Structures ====================== */

#define DDSD_CAPS        0x00000001 
#define DDSD_HEIGHT      0x00000002 
#define DDSD_WIDTH       0x00000004 
#define DDSD_PITCH       0x00000008 
#define DDSD_PIXELFORMAT 0x00001000 
#define DDSD_MIPMAPCOUNT 0x00020000 
#define DDSD_LINEARSIZE  0x00080000 
#define DDSD_DEPTH       0x00800000 

#define DDPF_ALPHAPIXELS 0x00000001 
#define DDPF_FOURCC      0x00000004 
#define DDPF_RGB         0x00000040 

#define DDSCAPS_COMPLEX  0x00000008 
#define DDSCAPS_TEXTURE  0x00001000 
#define DDSCAPS_MIPMAP   0x00400000 

#define DDS_MAGIC 0x20534444 /* "DDS " as a uint32 (little-endian) */

struct dds_pixelformat_st {
   uint32 size;                  /* Size of this structure in bytes (32) */
   uint32 flags;
   uint32 fourcc;
   uint32 rgb_bitcount;
   uint32 rgb_rmask;
   uint32 rgb_gmask;
   uint32 rgb_bmask;
   uint32 rgb_amask;
} __attribute__ ((packed));

struct dds_header_st {
   uint32 size;                  /* Size of this structure in bytes (124) */
   uint32 flags;
   uint32 height;
   uint32 width;
   uint32 linearsize;
   uint32 depth;
   uint32 mipmap_count;
   uint32 reserved1[11];
   struct dds_pixelformat_st pixelformat;
   uint32 caps1;
   uint32 caps2;
   uint32 reserved2[2];
   uint32 reserved3;
} __attribute__ ((packed));

#define FOURCC_DXT1 0x31545844 /* "DXT1" as a uint32 (little-endian) */
#define FOURCC_DXT2 0x32545844 /* "DXT2" as a uint32 (little-endian) */
#define FOURCC_DXT3 0x33545844 /* "DXT3" as a uint32 (little-endian) */
#define FOURCC_DXT4 0x34545844 /* "DXT4" as a uint32 (little-endian) */
#define FOURCC_DXT5 0x35545844 /* "DXT5" as a uint32 (little-endian) */

/* ------------------------- DXT Format Structures ------------------------ */

struct dxt_texel_color_st {
  uint16 color_0;
  uint16 color_1;
  uint8 pixel_color[4];
} __attribute__ ((packed));

struct dxt_texel_alpha_linear_st {
  uint16 pixel_alpha[4];
} __attribute__ ((packed));

struct dxt_texel_alpha_interp_st {
  uint8 alpha_0;
  uint8 alpha_1;
  uint8 pixel_alpha[6];
} __attribute__ ((packed));

/* ====================== TGA File Format Structures ===================== */

#define TGAFOOTER_SIGSTR          "TRUEVISION-XFILE."
#define TGA_CMAP_NONE             0    /* colormap_type for no colormap */
#define TGA_IMAGE_RGBA            2    /* Uncompressed, truecolor image */
#define TGA_ID_ORIENT_BOTTOM_LEFT 0x00 /* left-to-right, bottom-to-top */

struct tga_header_st {
   uint8 id_length;
   uint8 colormap_type;
   uint8 image_type;
   uint16 cmap_firstindex;
   uint16 cmap_length;
   uint8 cmap_entrysize;
   uint16 x_origin;
   uint16 y_origin;
   uint16 width;
   uint16 height;
   uint8 depth;
   uint8 image_descriptor;
} __attribute__ ((packed));

struct tga_footer_st {
   uint32 ext_offset;
   uint32 devel_offset;
   char signature[18];
} __attribute__ ((packed));

/* ====================== General Program Declarations ===================== */

struct imginfo_st {
  int width;
  int height;
  uint32 *buffer;
  uint32 fourcc;
};

//#define rgb565_to_rgba(x) ((((x) & 0xf800) << 8) | (((x) & 0x07e0) << 5) | (((x) & 0x001f) << 3) | 0xff000000)
//#define max(a,b) ((a) > (b) ? (a) : (b))

static inline int rgb565_to_rgba(int color) {
  return ((color & 0xf800) << 8) | ((color & 0x07e0) << 5) | ((color & 0x001f) << 3) | 0xff000000;
}

#ifdef max
#undef max
#endif
static inline int max(int a, int b) {
  return (a > b) ? a : b;
}

/* =========================== Begin Program Code ========================== */

int read_dds_headers(FILE *f, struct imginfo_st *imginfo) {
  struct dds_header_st dds_header;
  uint32 dds_magic;
  int required_flags;
  char fourccstr[5];

  if (fseek(f, 0, SEEK_SET)) {
    oserrmsg("Error seeking to beginning of input file");
    return 1;
  }
  if (fread(&dds_magic, sizeof(dds_magic), 1, f) != 1) {
    oserrmsg("Error reading DDS signature from input file");
    return 1;
  }
  if (dds_magic != DDS_MAGIC) {
    errmsg("Input file does not appear to be a DDS file");
    return 1;
  }
  if (fread(&dds_header, sizeof(dds_header), 1, f) != 1) {
    oserrmsg("Error reading DDS header from input file");
    return 1;
  }

  if (dds_header.size != sizeof(dds_header)) {
    errmsg("DDS file header is wrong size?  Unable to continue.");
    return 1;
  }
  required_flags = DDSD_CAPS | DDSD_PIXELFORMAT | DDSD_HEIGHT | DDSD_WIDTH
                 | DDSD_LINEARSIZE;
  if (((dds_header.flags & required_flags) != required_flags)) {
    /* According to Microsoft DDS documentation, we should also be checking for
     * DDSCAPS_TEXTURE and DDPF_FOURCC.  However, UnrealEd appears to
     * (incorrectly) write DDS files with the caps and most of the pixelformat
     * structure blank. */
    errmsg("Input file does not appear to contain a DXTC texture.");
    return 1;
  }
  switch (dds_header.pixelformat.fourcc) {
    case FOURCC_DXT1:
    case FOURCC_DXT3:
    case FOURCC_DXT5:
      break;
    default:
      memcpy(fourccstr, &dds_header.pixelformat.fourcc, 4);
      fourccstr[4] = 0;
      errmsg("%s compression type is not supported.", fourccstr);
      return 1;
  }
  
  imginfo->width = dds_header.width;
  imginfo->height = dds_header.height;
  imginfo->fourcc = dds_header.pixelformat.fourcc;

  return 0;
}

int read_decode_texel(FILE *f, struct imginfo_st *imginfo, uint32 *buf, int stride) {
  struct dxt_texel_color_st texel_color;
  struct dxt_texel_alpha_linear_st texel_alpha_linear;
  struct dxt_texel_alpha_interp_st texel_alpha_interp;
  uint32 color_map[4];
  uint8 alpha_map[8];
  uint32 alphadata[4][4];
  uint32 rgba;
  int i, j;
  uint32 pixel_alpha_01, pixel_alpha_23;
  uint8 *c0 = (uint8 *)(&color_map[0]);
  uint8 *c1 = (uint8 *)(&color_map[1]);
  uint8 *c2 = (uint8 *)(&color_map[2]);
  uint8 *c3 = (uint8 *)(&color_map[3]);

  if (imginfo->fourcc == FOURCC_DXT3) {
    if (fread(&texel_alpha_linear, sizeof(texel_alpha_linear), 1, f) != 1) {
      oserrmsg("Error reading image data from input file");
      return 1;
    }
    for (i=0; i<4; i++) {
      alphadata[i][0] = (texel_alpha_linear.pixel_alpha[i] << 4) & 0xf0;
      alphadata[i][1] = (texel_alpha_linear.pixel_alpha[i] << 0) & 0xf0;
      alphadata[i][2] = (texel_alpha_linear.pixel_alpha[i] >> 4) & 0xf0;
      alphadata[i][3] = (texel_alpha_linear.pixel_alpha[i] >> 8) & 0xf0;
    }
  } else if (imginfo->fourcc == FOURCC_DXT5) {
    if (fread(&texel_alpha_interp, sizeof(texel_alpha_interp), 1, f) != 1) {
      oserrmsg("Error reading image data from input file");
      return 1;
    }
    alpha_map[0] = texel_alpha_interp.alpha_0; // bit code 000
    alpha_map[1] = texel_alpha_interp.alpha_1; // bit code 001
    if (alpha_map[0] <= alpha_map[1]) {
      alpha_map[2] = (4*alpha_map[0] + 1*alpha_map[1] + 2) / 5; // Bit code 010
      alpha_map[3] = (3*alpha_map[0] + 2*alpha_map[1] + 2) / 5; // Bit code 011
      alpha_map[4] = (2*alpha_map[0] + 3*alpha_map[1] + 2) / 5; // Bit code 100
      alpha_map[5] = (1*alpha_map[0] + 4*alpha_map[1] + 2) / 5; // Bit code 101
      alpha_map[6] = 0;                                         // Bit code 110
      alpha_map[7] = 255;                                       // Bit code 111
    } else {
      alpha_map[2] = (6*alpha_map[0] + 1*alpha_map[1] + 3) / 7; // bit code 010
      alpha_map[3] = (5*alpha_map[0] + 2*alpha_map[1] + 3) / 7; // bit code 011
      alpha_map[4] = (4*alpha_map[0] + 3*alpha_map[1] + 3) / 7; // bit code 100
      alpha_map[5] = (3*alpha_map[0] + 4*alpha_map[1] + 3) / 7; // bit code 101
      alpha_map[6] = (2*alpha_map[0] + 5*alpha_map[1] + 3) / 7; // bit code 110
      alpha_map[7] = (1*alpha_map[0] + 6*alpha_map[1] + 3) / 7; // bit code 111
    }
    pixel_alpha_01 = (texel_alpha_interp.pixel_alpha[2] << 16) |
                     (texel_alpha_interp.pixel_alpha[1] << 8) |
                     (texel_alpha_interp.pixel_alpha[0]);
    pixel_alpha_23 = (texel_alpha_interp.pixel_alpha[5] << 16) |
                     (texel_alpha_interp.pixel_alpha[4] << 8) |
                     (texel_alpha_interp.pixel_alpha[3]);
    alphadata[0][0] = alpha_map[(pixel_alpha_01 >>  0) & 0x07];
    alphadata[0][1] = alpha_map[(pixel_alpha_01 >>  3) & 0x07];
    alphadata[0][2] = alpha_map[(pixel_alpha_01 >>  6) & 0x07];
    alphadata[0][3] = alpha_map[(pixel_alpha_01 >>  9) & 0x07];
    alphadata[1][0] = alpha_map[(pixel_alpha_01 >> 12) & 0x07];
    alphadata[1][1] = alpha_map[(pixel_alpha_01 >> 15) & 0x07];
    alphadata[1][2] = alpha_map[(pixel_alpha_01 >> 18) & 0x07];
    alphadata[1][3] = alpha_map[(pixel_alpha_01 >> 21) & 0x07];
    alphadata[2][0] = alpha_map[(pixel_alpha_23 >>  0) & 0x07];
    alphadata[2][1] = alpha_map[(pixel_alpha_23 >>  3) & 0x07];
    alphadata[2][2] = alpha_map[(pixel_alpha_23 >>  6) & 0x07];
    alphadata[2][3] = alpha_map[(pixel_alpha_23 >>  9) & 0x07];
    alphadata[3][0] = alpha_map[(pixel_alpha_23 >> 12) & 0x07];
    alphadata[3][1] = alpha_map[(pixel_alpha_23 >> 15) & 0x07];
    alphadata[3][2] = alpha_map[(pixel_alpha_23 >> 18) & 0x07];
    alphadata[3][3] = alpha_map[(pixel_alpha_23 >> 21) & 0x07];
  } else {
    for (i=0; i<4; i++) {
      for (j=0; j<4; j++) {
        alphadata[i][j] = 0xff;
      }
    }
  }
  for (i=0; i<4; i++) {
    for (j=0; j<4; j++) {
      alphadata[i][j] = (alphadata[i][j] << 24) | 0x00ffffff;
    }
  }

  if (fread(&texel_color, sizeof(texel_color), 1, f) != 1) {
    oserrmsg("Error reading image data from input file");
    return 1;
  }
  color_map[0] = rgb565_to_rgba(texel_color.color_0);
  color_map[1] = rgb565_to_rgba(texel_color.color_1);

  if ((imginfo->fourcc == FOURCC_DXT1) &&
      (texel_color.color_0 <= texel_color.color_1)) {
    for (i=0; i<4; i++) {
      c2[i] = (c0[i] + c1[i]) / 2;
    }
    color_map[3] = 0x00000000;
  } else {
    for (i=0; i<4; i++) {
      c2[i] = ((2 * c0[i]) + (1 * c1[i]) + 1) / 3;
      c3[i] = ((1 * c0[i]) + (2 * c1[i]) + 1) / 3;
    }
  }

  for (i=0; i<4; i++) {
    for (j=0; j<4; j++) {
      rgba = color_map[(texel_color.pixel_color[i] >> (j*2)) & 0x03]
            & alphadata[i][j];
      buf[(i * stride) + j] = rgba;
    }
  }

  return 0;
}

int read_dxt_imagedata(FILE *f, struct imginfo_st *imginfo) {
  int i, j;
  int texels_per_row = max(imginfo->width >> 2, 1);
  int texel_rows = max(imginfo->height >> 2, 1);
  int stride = imginfo->width;

  for (i = 0; i < texel_rows; i++) {
    for (j = 0; j < texels_per_row; j++) {
      if (read_decode_texel(f, imginfo, &(imginfo->buffer[(i*4*stride)+j*4]), stride)) {
        return 1;
      }
    }
  }

  return 0;
}

int write_tga_headers(FILE *f, struct imginfo_st *imginfo) {
  struct tga_header_st tga_header = {
    0, TGA_CMAP_NONE, TGA_IMAGE_RGBA, 0, 0, 0, 0, 0,
    imginfo->width, imginfo->height, 32,
    TGA_ID_ORIENT_BOTTOM_LEFT | 8
  };

  if (fwrite(&tga_header, sizeof(tga_header), 1, f) != 1) {
    oserrmsg("Error writing TGA header to output file");
    return 1;
  }

  return 0;
}

int write_tga_imagedata(FILE *f, struct imginfo_st *imginfo) {
  int row;
  int stride = imginfo->width;

  for (row = imginfo->height; row >= 0; row--) {
    if (fwrite(&(imginfo->buffer[row*stride]), imginfo->width * 4, 1, f) != 1) {
      oserrmsg("Error writing image data to output file");
      return 1;
    }
  }

  return 0;
}

int write_tga_footer(FILE *f, struct imginfo_st *imginfo) {
  struct tga_footer_st tga_footer = {
    0, 0, TGAFOOTER_SIGSTR
  };

  if (fwrite(&tga_footer, sizeof(tga_footer), 1, f) != 1) {
    oserrmsg("Error writing TGA footer to output file");
    return 1;
  }

  return 0;
}

char *progname;

void usage(void) {
  errmsg("Usage: %s inputfile.dds outputfile.tga", progname);
}

int main(int argc, char *argv[]) {
  char *infile_name = NULL;
  char *outfile_name = NULL;
  FILE *infile;
  FILE *outfile;
  struct imginfo_st imginfo;
  int i;

  progname = argv[0];

  for (i=1; i<argc; i++) {
    if (!infile_name) {
      infile_name = argv[i];
    } else if (!outfile_name) {
      outfile_name = argv[i];
    } else {
      usage(); return 1;
    }
  }

  if (!infile_name) {
    usage(); return 1;
  }

  infile = fopen(infile_name, "rb");
  if (!infile) {
    oserrmsg("Error opening input file");
    return 2;
  }

  if (!outfile_name) {
#ifndef WIN32_GUI
    usage(); return 1;
#else
    char *ext;
    char *outfile_ext;
    DWORD result;

    memset(&ofn, sizeof(ofn), 0);
    outfile_name = malloc(PATH_MAX);

    outfile_ext = ".tga";
    ofn.lpstrFilter = "TGA Files (*.tga)\0*.tga\0All Files (*.*)\0*.*\0";

    ofn.lStructSize = sizeof(ofn);
    ofn.lpstrFile = outfile_name;
    ofn.nMaxFile = PATH_MAX;
    ofn.Flags = OFN_OVERWRITEPROMPT;
    ofn.lpstrDefExt = outfile_ext+1;

    strncpy(outfile_name, infile_name, PATH_MAX-5);
    ext = strrchr(outfile_name, '.');
    if (!ext ||
        (strcasecmp(ext, ".dds"))) {
      ext = outfile_name + strlen(outfile_name);
    }
    strcpy(ext, outfile_ext);

    result = GetSaveFileName(&ofn);
    if (!result) {
      result = CommDlgExtendedError();
      if (result) {
        errmsg("Save dialog error: 0x%lx", result);
        return 9;
      } else {
        return 0;
      }
    }
#endif
  }

  if (read_dds_headers(infile, &imginfo)) {
    return 3;
  }
  imginfo.buffer = malloc(imginfo.width * imginfo.height * sizeof(uint32));
  if (!imginfo.buffer) {
    errmsg("Unable to allocate memory for buffer");
    return 4;
  }
  if (read_dxt_imagedata(infile, &imginfo)) {
    return 5;
  }

  outfile = fopen(outfile_name, "wb");
  if (!outfile) {
    oserrmsg("Error opening output file");
    return 6;
  }
  if (write_tga_headers(outfile, &imginfo)) {
    return 7;
  }
  if (write_tga_imagedata(outfile, &imginfo)) {
    return 8;
  }
  if (write_tga_footer(outfile, &imginfo)) {
    return 9;
  }
  
  usermsg("Conversion to TGA complete.");

  return 0;
}
