/* g16convert - Convert UnrealEd G16 textures in (broken) BMP form to/from TIFF
 * Author: Alex Stewart <alex@foogod.com>
 * Version: 1.0
 * Date: Dec 3, 2004
 * 
 * This code has been placed into the public domain by the author.
 *
 * Description:
 *
 * UnrealEd (the map editor for the series of games based on the Unreal Engine)
 * has an internal "G16" (16-bit grayscale) image representation which it uses
 * for such things as terrain heightmaps.  It has the ability to import and
 * export these images from/to files, but only supports importing/exporting
 * them as 16-bit grayscale BMP files.
 *
 * The problem is that there's no such thing as a 16-bit grayscale BMP file.
 * Such images are not supported by the file format.  Not only does UnrealEd
 * write the data in a nonstandard format, but it writes it in a way that is
 * indistinguishable from real 16-bit color (RGB) BMP files, thus creating a
 * BMP file which cannot be loaded properly (as a grayscale image) by any
 * standards-conforming graphics program, even if it wanted to.
 *
 * This program is designed to convert the (broken) G16-BMP files created by
 * UnrealEd into (standard) 16-bit grayscale TIFF images (why they didn't just
 * use TIFF in the first place I don't understand), and can also convert them
 * back again for easy importing back into UnrealEd after editing.
 *
 * To use this program, simply invoke it with an input filename and an output
 * filename.  It will automatically detect whether the input file is G16 or
 * TIFF and will convert appropriately.
 *
 * When converting from G16 to TIFF, you can optionally give the "-8"
 * ("--8bit") command line switch, which will write an 8-bit TIFF file instead
 * of a 16-bit one (in case you don't have any 16-bit capable graphics programs
 * available).  Going the other way isn't currently supported (UnrealEd can do
 * this itself internally from the terrain editor window if you just import
 * from an 8-bit BMP file or equivalent).
 */

/* This program does not currently handle converting any formats other than
   16-bit uncompressed grayscale TIFF into G16 format.  Any 16-bit capable
   graphics programs out there should be able to save in this format, and for
   non-16-bit-capable stuff, there are plenty of other utilities for converting
   8-bit formats into such things as 8-bit BMP (UnrealEd can convert 8-bit
   images to G16 internally once imported).  This program is mostly intended to
   be useful if you want to preserve the full 16-bit information through the
   external editing process, but want to use standard tools.
 */

/* Note:  This code assumes a little-endian (Intel) machine architecture.  I 
   could make it architecture-independent, but it just doesn't seem worth the
   effort since as far as I know UnrealEd is only available for x86 anyway.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>

#define WINDOW_TITLE "G16/TIFF Conversion"

#ifndef WIN32_GUI

/* We're compiling the normal command-line version.  Have messages go to
 * stdout/stderr */

#define infomsg(fmt, args...) { printf(fmt "\n", ## args); fflush(stdout); }
#define usermsg(fmt, args...) { printf(fmt "\n", ## args); fflush(stdout); }
#define errmsg(fmt, args...) fprintf(stderr, "Error: " fmt "\n", ## args)
#define oserrmsg(str) perror(str)

#else

/* We're compiling this for the "_w" (Windows GUI mode) executable.  Do user
 * interaction via dialog boxes and such */

#include <windows.h>

#define infomsg(fmt, args...) /* Don't show info messages in GUI mode */
#define usermsg(fmt, args...) dialog_box(MB_OK | MB_ICONINFORMATION, fmt, ## args)
#define errmsg(fmt, args...) dialog_box(MB_OK | MB_ICONERROR, fmt, ## args)
#define oserrmsg(str) dialog_box(MB_OK | MB_ICONERROR, "%s: %s", str, strerror(errno))

const char window_title[] = WINDOW_TITLE;

#define MSGBOX_BUFSIZE 1024

int dialog_box(int flags, const char *fmt, ...) {
  va_list ap;
  char buf[MSGBOX_BUFSIZE];

  va_start(ap, fmt);
  vsnprintf(buf, MSGBOX_BUFSIZE, fmt, ap);
  return MessageBox(NULL, buf, window_title, flags);
}

/* This is for the GetSaveFileName call in main().  For some reason, it doesn't
 * like it if this structure is on the stack, so we'll put it here instead.  I
 * still don't really understand it, but that's Windows for ya.. */

OPENFILENAME ofn;

#endif

typedef unsigned int uint32;
typedef unsigned short uint16;

#define swap16(x) ( (((x) & 0x00ff) << 8) | (((x) & 0xff00) >> 8) )
#define swap32(x) ( (((x) & 0x000000ff) << 24) | (((x) & 0x0000ff00) << 8) | \
                    (((x) & 0x00ff0000) >> 8)  | (((x) & 0xff000000) >> 24) )

#define swapbits16(x) ( (((uint16)bitswap_table[((x) & 0xff00) >> 8]) << 8) | \
                        bitswap_table[(x) & 0x00ff] )

unsigned char bitswap_table[256] = {
0x00,0x80,0x40,0xc0,0x20,0xa0,0x60,0xe0,0x10,0x90,0x50,0xd0,0x30,0xb0,0x70,0xf0,
0x08,0x88,0x48,0xc8,0x28,0xa8,0x68,0xe8,0x18,0x98,0x58,0xd8,0x38,0xb8,0x78,0xf8,
0x04,0x84,0x44,0xc4,0x24,0xa4,0x64,0xe4,0x14,0x94,0x54,0xd4,0x34,0xb4,0x74,0xf4,
0x0c,0x8c,0x4c,0xcc,0x2c,0xac,0x6c,0xec,0x1c,0x9c,0x5c,0xdc,0x3c,0xbc,0x7c,0xfc,
0x02,0x82,0x42,0xc2,0x22,0xa2,0x62,0xe2,0x12,0x92,0x52,0xd2,0x32,0xb2,0x72,0xf2,
0x0a,0x8a,0x4a,0xca,0x2a,0xaa,0x6a,0xea,0x1a,0x9a,0x5a,0xda,0x3a,0xba,0x7a,0xfa,
0x06,0x86,0x46,0xc6,0x26,0xa6,0x66,0xe6,0x16,0x96,0x56,0xd6,0x36,0xb6,0x76,0xf6,
0x0e,0x8e,0x4e,0xce,0x2e,0xae,0x6e,0xee,0x1e,0x9e,0x5e,0xde,0x3e,0xbe,0x7e,0xfe,
0x01,0x81,0x41,0xc1,0x21,0xa1,0x61,0xe1,0x11,0x91,0x51,0xd1,0x31,0xb1,0x71,0xf1,
0x09,0x89,0x49,0xc9,0x29,0xa9,0x69,0xe9,0x19,0x99,0x59,0xd9,0x39,0xb9,0x79,0xf9,
0x05,0x85,0x45,0xc5,0x25,0xa5,0x65,0xe5,0x15,0x95,0x55,0xd5,0x35,0xb5,0x75,0xf5,
0x0d,0x8d,0x4d,0xcd,0x2d,0xad,0x6d,0xed,0x1d,0x9d,0x5d,0xdd,0x3d,0xbd,0x7d,0xfd,
0x03,0x83,0x43,0xc3,0x23,0xa3,0x63,0xe3,0x13,0x93,0x53,0xd3,0x33,0xb3,0x73,0xf3,
0x0b,0x8b,0x4b,0xcb,0x2b,0xab,0x6b,0xeb,0x1b,0x9b,0x5b,0xdb,0x3b,0xbb,0x7b,0xfb,
0x07,0x87,0x47,0xc7,0x27,0xa7,0x67,0xe7,0x17,0x97,0x57,0xd7,0x37,0xb7,0x77,0xf7,
0x0f,0x8f,0x4f,0xcf,0x2f,0xaf,0x6f,0xef,0x1f,0x9f,0x5f,0xdf,0x3f,0xbf,0x7f,0xff,
};

/* ====================== BMP File Format Structures ====================== */

struct bmp_header_st {
   uint16 magic;                 /* Magic identifier ("BM")     */
   uint32 size;                  /* File size in bytes          */
   uint16 reserved1;
   uint16 reserved2;
   uint32 data_offset;           /* File offset of image data */
} __attribute__ ((packed));

struct bmp_info_st {
   uint32 header_size;           /* Header size in bytes */
   uint32 width;                 /* Width in pixels */
   uint32 height;                /* Height in pixels */
   uint16 planes;                /* Number of color planes */
   uint16 bits;                  /* Bits per pixel */
   uint32 compression;           /* Compression type */
   uint32 image_size;            /* Image size in bytes */
   uint32 xres;                  /* X resolution (pixels per meter) */
   uint32 yres;                  /* Y resolution (pixels per meter) */
   uint32 used_colors;           /* Number of colors used (colortable size) */
   uint32 important_colors;      /* Number of important colors */
} __attribute__ ((packed));

/* BMP file magic signature value */

#define BMP_MAGIC 0x4d42 /* 'MB' ("BM" signature byte-swapped) */

/* Relevant values for compression */

#define BMP_COMPR_RGB       0 /* No compression */
#define BMP_COMPR_BITFIELDS 3 /* No compression, color table contains component masks */

/* ====================== TIFF File Format Structures ===================== */

struct tiff_header_st {
  uint16 byte_order;  /* Byte order indicator ('II' or 'MM') */
  uint16 magic;       /* TIFF file signature (0x002a, byte-order-dependent) */
  uint32 ifd_offset;  /* File offset of first IFD table */
} __attribute__ ((packed));

struct tiff_ifd_entry_st {
  uint16 tag;       /* Tag identifying what the value means */
  uint16 type;      /* Type of value(s) */
  uint32 count;     /* Number of values */
  uint32 value;     /* Value or pointer to value(s) depending on type/count */
} __attribute__ ((packed));

/* Format of "rational" datatype */
struct tiff_rational_st {
  uint32 numerator;
  uint32 denominator;
} __attribute__ ((packed));

/* Valid TIFF byte-order field values */

#define TIFF_BYTEORDER_LITTLE 0x4949 /* 'II' */
#define TIFF_BYTEORDER_BIG    0x4d4d /* 'MM' */

/* TIFF file magic signature value */

#define TIFF_MAGIC 0x002a /* NOTE: This is subject to file byte-order */

/* Relevant TIFF value types: */

#define TIFF_SHORT 3
#define TIFF_LONG 4
#define TIFF_RATIONAL 5

/* Relevant TIFF tags: */

#define TIFF_TAG_WIDTH    0x100 /* ImageWidth */
#define TIFF_TAG_LENGTH   0x101 /* ImageLength */
#define TIFF_TAG_BITS     0x102 /* BitsPerSample */
#define TIFF_TAG_COMPR    0x103 /* Compression */
#define TIFF_TAG_PMI      0x106 /* PhotometricInterpretation */
#define TIFF_TAG_FILLORD  0x10a /* FillOrder */
#define TIFF_TAG_STROFFS  0x111 /* StripOffsets */
#define TIFF_TAG_ORIENT   0x112 /* Orientation */
#define TIFF_TAG_STRROWS  0x116 /* RowsPerStrip */
#define TIFF_TAG_STRBYTES 0x117 /* StripByteCounts */
#define TIFF_TAG_XRES     0x11a /* XResolution */
#define TIFF_TAG_YRES     0x11b /* YResolution */
#define TIFF_TAG_RESUNIT  0x128 /* ResolutionUnit */

/* Relevant values for Compression */

#define TIFF_COMPR_NONE 1  /* No compression used */

/* Relevant values for PhotometricInterpretation */

#define TIFF_PMI_WHITEZERO 0  /* WhiteIsZero */
#define TIFF_PMI_BLACKZERO 1  /* BlackIsZero */

/* Relevant values for FillOrder */

#define TIFF_FILL_NORM  1 /* msb-to-lsb (for our purposes, "normal") */
#define TIFF_FILL_WRONG 2 /* lsb-to-msb (bits are inverted within the byte) */

/* Relevant values for Orientation */

#define TIFF_ORIENT_TOP_LEFT    1 /* top-left (default) */
#define TIFF_ORIENT_BOTTOM_LEFT 4 /* bottom-left */

/* ====================== General Program Declarations ===================== */

struct imginfo_st {
  int width;
  int height;
  uint16 *buffer;
  int convert_depth;
  /* for BMP: */
  uint32 data_offset;
  /* for TIFF: */
  int byteswap;
  int bitswap;
  int invert;
  int bottom_up;
  int rowsperstrip;
  int numstrips;
  uint32 *strips;
};

#define FILETYPE_UNKNOWN 0
#define FILETYPE_BMP     1
#define FILETYPE_TIFF    2

/* =========================== Begin Program Code ========================== */

int identify_file(FILE *f) {
  uint16 magic;

  if (fread(&magic, 2, 1, f) != 1) {
    oserrmsg("Unable to read from input file");
    return -1;
  }

  switch (magic) {
    case BMP_MAGIC:             return FILETYPE_BMP;
    case TIFF_BYTEORDER_LITTLE: return FILETYPE_TIFF;
    case TIFF_BYTEORDER_BIG:    return FILETYPE_TIFF;
  }
  return FILETYPE_UNKNOWN;
}

int read_tiff_headers(FILE *f, struct imginfo_st *imginfo) {
  struct tiff_header_st tiff_header;
  struct tiff_ifd_entry_st *ifd_entries;
  int byteswap;
  uint16 ifd_count;
  uint32 ifd_offset;
  int ifdentsize = sizeof(struct tiff_ifd_entry_st);
  int i;
  struct tiff_ifd_entry_st *entry;
  int orientation, bits_per_sample, compression, photometric_interp;

  if (fseek(f, 0, SEEK_SET)) {
    oserrmsg("Error seeking to beginning of input file");
    return 1;
  }
  if (fread(&tiff_header, sizeof(tiff_header), 1, f) != 1) {
    oserrmsg("Error reading TIFF header from input file");
    return 1;
  }
  if (tiff_header.byte_order == TIFF_BYTEORDER_BIG) {
    byteswap = 1;
    ifd_offset = swap32(tiff_header.ifd_offset);
  } else {
    byteswap = 0;
    ifd_offset = tiff_header.ifd_offset;
  }
  if (fseek(f, ifd_offset, SEEK_SET)) {
    oserrmsg("Error locating IFD header in input file");
    return 1;
  }
  if (fread(&ifd_count, sizeof(ifd_count), 1, f) != 1) {
    oserrmsg("Error reading IFD header from input file");
    return 1;
  }
  if (byteswap) ifd_count = swap16(ifd_count);
  ifd_entries = (struct tiff_ifd_entry_st *)malloc(ifdentsize * ifd_count);
  if (!ifd_entries) {
    errmsg("Error allocating memory for IFD header");
    return 1;
  }
  if (fread(ifd_entries, ifdentsize, ifd_count, f) != ifd_count) {
    oserrmsg("Error reading IFD entries from input file");
    return 1;
  }

  imginfo->byteswap = byteswap;
  imginfo->bitswap = 0;

  imginfo->width = 0;
  imginfo->height = 0;
  imginfo->strips = 0;
  imginfo->rowsperstrip = 0;
  orientation = TIFF_ORIENT_TOP_LEFT;
  bits_per_sample = 0;
  compression = 0;
  photometric_interp = 0;

  for (i=0; i<ifd_count; i++) {
    entry = &ifd_entries[i];
    if (byteswap) {
      entry->tag = swap16(entry->tag);
      entry->type = swap32(entry->type);
      entry->count = swap32(entry->count);
      entry->value = swap32(entry->value);
    }
    switch (entry->tag) {
      case TIFF_TAG_WIDTH: imginfo->width = entry->value; break;
      case TIFF_TAG_LENGTH: imginfo->height = entry->value; break;
      case TIFF_TAG_STROFFS:
        imginfo->numstrips = entry->count;
        imginfo->strips = (uint32 *)malloc(sizeof(uint32) * entry->count);
        if (!imginfo->strips) {
          errmsg("Error allocating memory for strip table");
          return 1;
        }
        if (entry->count == 1) {
          imginfo->strips[0] = entry->value;
        } else {
          if (fseek(f, entry->value, SEEK_SET)) {
            oserrmsg("Error locating strip table in input file");
            return 1;
          }
          if (fread(imginfo->strips, sizeof(uint32), entry->count, f) != entry->count) {
            oserrmsg("Error reading strip table from input file");
            return 1;
          }
          if (byteswap) {
            for (i=0; i<entry->count; i++) {
              imginfo->strips[i] = swap32(imginfo->strips[i]);
            }
          }
        }
        break;
      case TIFF_TAG_STRROWS: imginfo->rowsperstrip = entry->value; break;
      case TIFF_TAG_FILLORD:
	/* GRUMBLE:  This tag shouldn't even be meaningful for 16-bit files,
           but ImageMagick apparently writes 16-bit grayscale files with
	   an inverted bit order and sets "FillOrder" to "2" to indicate this.
	   This is quite clearly WRONG behavior according to the TIFF spec
	   (FillOrder only applies to _pixel_ order within bytes, not bit order
           within a (part of a) pixel), but I guess we need to support it. */
        imginfo->bitswap = (entry->value == TIFF_FILL_WRONG);
        break;
      case TIFF_TAG_ORIENT: orientation = entry->value; break;
      case TIFF_TAG_BITS: bits_per_sample = entry->value; break;
      case TIFF_TAG_COMPR: compression = entry->value; break;
      case TIFF_TAG_PMI: photometric_interp = entry->value; break;
    }
  }

  imginfo->invert = (photometric_interp == TIFF_PMI_WHITEZERO);
  imginfo->bottom_up = (orientation == TIFF_ORIENT_BOTTOM_LEFT);

  if (!imginfo->height || !imginfo->width) {
    errmsg("Bad height/width data in input file.  Aborting.");
    return 1;
  }
  if (!imginfo->strips || !imginfo->rowsperstrip ||
      (imginfo->numstrips <= (imginfo->height-1) / imginfo->rowsperstrip)) {
    errmsg("Bad or missing image data pointers.  Aborting.");
    return 1;
  }
  if ((bits_per_sample != 16) || (compression != TIFF_COMPR_NONE)) {
    errmsg("Input is not a 16-bit uncompressed TIFF file.  Aborting.");
    return 1;
  }
  if ((photometric_interp != TIFF_PMI_BLACKZERO) &&
      (photometric_interp != TIFF_PMI_WHITEZERO)) {
    errmsg("Input is apparently not a grayscale file.  Unable to Continue.");
    return 1;
  }
  if ((orientation != TIFF_ORIENT_TOP_LEFT) &&
      (orientation != TIFF_ORIENT_BOTTOM_LEFT)) {
    errmsg("Unsupported image orientation in TIFF file.  Unable to Continue.");
    return 1;
  }

  imginfo->buffer = (uint16 *)malloc(imginfo->width * 2);
  if (!imginfo->buffer) {
    errmsg("Error allocating image data buffer");
    return 1;
  }

  return 0;
}

int write_tiff_headers(FILE *f, struct imginfo_st *imginfo) {
  int w = imginfo->width;
  int h = imginfo->height;
  int depth = imginfo->convert_depth ? 8 : 16;
  int imgsize = w * h * (depth >> 3);
  struct tiff_header_st tiff_header = {
    TIFF_BYTEORDER_LITTLE,
    TIFF_MAGIC,
    sizeof(struct tiff_header_st)
  };
  /* NOTE: The following entries must be in numerical tag order! */
  struct tiff_ifd_entry_st ifd_entries[] = {
    { TIFF_TAG_WIDTH,    TIFF_SHORT, 1, w},
    { TIFF_TAG_LENGTH,   TIFF_SHORT, 1, h},
    { TIFF_TAG_BITS,     TIFF_SHORT, 1, depth},   /* (8/16 bits per sample) */
    { TIFF_TAG_COMPR,    TIFF_SHORT, 1, TIFF_COMPR_NONE},
    { TIFF_TAG_PMI,      TIFF_SHORT, 1, TIFF_PMI_BLACKZERO},
    { TIFF_TAG_STROFFS,  TIFF_LONG,  1, 0},    /* (we fill this in below) */
    { TIFF_TAG_STRROWS,  TIFF_SHORT, 1, h},    /* (whole image in 1 strip) */
    { TIFF_TAG_STRBYTES, TIFF_LONG,  1, imgsize},/* (strip = whole image) */
    { TIFF_TAG_XRES,  TIFF_RATIONAL, 1, 0},    /* (we fill this in below) */
    { TIFF_TAG_YRES,  TIFF_RATIONAL, 1, 0},    /* (we fill this in below) */
    { TIFF_TAG_RESUNIT,  TIFF_SHORT, 1, 1},    /* (no meaningful units) */
  };
  uint16 ifd_count = sizeof(ifd_entries) / sizeof(ifd_entries[0]);
  uint32 zero = 0;
  struct tiff_rational_st xres = {1, 1};
  struct tiff_rational_st yres = {1, 1};

  ifd_entries[8].value = sizeof(tiff_header) + sizeof(ifd_count) + sizeof(ifd_entries) + sizeof(zero);
  ifd_entries[9].value = ifd_entries[8].value + sizeof(xres);
  ifd_entries[5].value = ifd_entries[9].value + sizeof(yres);

  if (fwrite(&tiff_header, sizeof(tiff_header), 1, f) != 1) {
    oserrmsg("Error writing TIFF header to output file");
    return 1;
  }
  if (fwrite(&ifd_count, sizeof(ifd_count), 1, f) != 1) {
    oserrmsg("Error writing IFD header to output file");
    return 1;
  }
  if (fwrite(&ifd_entries, sizeof(ifd_entries), 1, f) != 1) {
    oserrmsg("Error writing IFD entries to output file");
    return 1;
  }
  if (fwrite(&zero, sizeof(zero), 1, f) != 1) {
    oserrmsg("Error writing IFD terminator to output file");
    return 1;
  }
  if (fwrite(&xres, sizeof(xres), 1, f) != 1) {
    oserrmsg("Error writing resolution data to output file");
    return 1;
  }
  if (fwrite(&yres, sizeof(yres), 1, f) != 1) {
    oserrmsg("Error writing resolution data to output file");
    return 1;
  }

  return 0;
}

int read_tiff_row(FILE *f, struct imginfo_st *imginfo, int lineno) {
  int stripnum = lineno / imginfo->rowsperstrip;
  int stripline = lineno % imginfo->rowsperstrip;
  int i;

  if (fseek(f, imginfo->strips[stripnum] + (stripline * (imginfo->width * 2)), SEEK_SET)) {
    oserrmsg("Error locating image data in input file");
    return 1;
  }
  if (fread(imginfo->buffer, 2, imginfo->width, f) != imginfo->width) {
    oserrmsg("Error reading image data from input file");
    return 1;
  }
  if (imginfo->byteswap) {
    for (i=0; i<imginfo->width; i++) {
      imginfo->buffer[i] = swap16(imginfo->buffer[i]);
    }
  }
  if (imginfo->bitswap) {
    /* GRUMBLE: ImageMagick is broken.  See comments under read_tiff_headers */
    for (i=0; i<imginfo->width; i++) {
      imginfo->buffer[i] = swapbits16(imginfo->buffer[i]);
    }
  }
  if (imginfo->invert) {
    for (i=0; i<imginfo->width; i++) {
      imginfo->buffer[i] = 0xffff - imginfo->buffer[i];
    }
  }

  return 0;
}

int write_tiff_row(FILE *f, struct imginfo_st *imginfo) {
  int i;

  if (imginfo->convert_depth) {
    for (i=0; i<imginfo->width; i++) {
      /* Write the MSB of each value */
      if (fwrite(((char *)&imginfo->buffer[i])+1, 1, 1, f) != 1) {
        oserrmsg("Error writing image data to output file");
        return 1;
      }
    }
  } else {
    if (fwrite(imginfo->buffer, 2, imginfo->width, f) != imginfo->width) {
      oserrmsg("Error writing image data to output file");
      return 1;
    }
  }
  return 0;
}

int read_bmp_headers(FILE *f, struct imginfo_st *imginfo) {
  struct bmp_header_st bmp_header;
  struct bmp_info_st info_header;

  if (fseek(f, 0, SEEK_SET)) {
    oserrmsg("Error seeking to beginning of input file");
    return 1;
  }
  if (fread(&bmp_header, sizeof(bmp_header), 1, f) != 1) {
    oserrmsg("Error reading BMP header from input file");
    return 1;
  }
  if (fread(&info_header, sizeof(info_header), 1, f) != 1) {
    oserrmsg("Error reading info header from input file");
    return 1;
  }

  if ((bmp_header.magic != BMP_MAGIC) || /* shouldn't happen */
      (info_header.bits != 16) ||
      (info_header.planes != 1) ||
      ((info_header.compression != BMP_COMPR_RGB) &&
       (info_header.compression != BMP_COMPR_BITFIELDS)) ) {
    errmsg("Input file does not appear to be in 16-bit raw BMP format.  Aborting.");
    return 1;
  }

  imginfo->width = info_header.width;
  imginfo->height = info_header.height;
  imginfo->data_offset = bmp_header.data_offset;

  imginfo->buffer = (uint16 *)malloc(imginfo->width * 2);
  if (!imginfo->buffer) {
    errmsg("Error allocating image data buffer");
    return 1;
  }

  return 0;
}

int write_bmp_headers(FILE *f, struct imginfo_st *imginfo) {
  int data_offset = sizeof(struct bmp_header_st) + sizeof(struct bmp_info_st);
  int img_size = imginfo->height * imginfo->width * 2;
  int file_size = data_offset + img_size;
  struct bmp_header_st bmp_header = {BMP_MAGIC, file_size, 0, 0, data_offset};
  struct bmp_info_st info_header = {
    sizeof(struct bmp_info_st),  /* Header size */
    imginfo->width,              /* Width */
    imginfo->height,             /* Height */
    1,                           /* Color planes */
    16,                          /* Bits per pixel */
    BMP_COMPR_RGB,               /* Compression (none) */
    img_size,                    /* Size of image data */
    0,                           /* X resolution */
    0,                           /* Y resolution */
    0,                           /* Colors used */
    0,                           /* Important colors */
  };

  if (fwrite(&bmp_header, sizeof(bmp_header), 1, f) != 1) {
    oserrmsg("Error writing BMP header to output file");
    return 1;
  }
  if (fwrite(&info_header, sizeof(info_header), 1, f) != 1) {
    oserrmsg("Error writing info header to output file");
    return 1;
  }

  return 0;
}

int read_g16_row(FILE *f, struct imginfo_st *imginfo, int lineno) {
  if (fseek(f, imginfo->data_offset + (lineno * imginfo->width * 2), SEEK_SET)) {
    oserrmsg("Error locating image data in input file");
    return 1;
  }
  if (fread(imginfo->buffer, 2, imginfo->width, f) != imginfo->width) {
    oserrmsg("Error reading image data from input file");
    return 1;
  }

  return 0;
}

int write_g16_row(FILE *f, struct imginfo_st *imginfo) {
  if (fwrite(imginfo->buffer, 2, imginfo->width, f) != imginfo->width) {
    oserrmsg("Error writing image data to output file");
    return 1;
  }
  return 0;
}

char *progname;

void usage(void) {
  errmsg("Usage: %s [-8] inputfile.bmp outputfile.tiff\n -or-  %s inputfile.tiff outputfile.bmp", progname, progname);
}

int main(int argc, char *argv[]) {
  char *infile_name = NULL;
  char *outfile_name = NULL;
  FILE *infile;
  FILE *outfile;
  int filetype;
  struct imginfo_st imginfo;
  int i, row;

  progname = argv[0];

  imginfo.convert_depth = 0;

  for (i=1; i<argc; i++) {
    if (!strcmp(argv[i], "-8") || !strcmp(argv[i], "--8bit")) {
      imginfo.convert_depth = 1;
    } else if (!infile_name) {
      infile_name = argv[i];
    } else if (!outfile_name) {
      outfile_name = argv[i];
    } else {
      usage(); return 1;
    }
  }

  if (!infile_name) {
    usage(); return 1;
  }

  infile = fopen(infile_name, "rb");
  if (!infile) {
    oserrmsg("Error opening input file");
    return 2;
  }

  filetype = identify_file(infile);

  if (filetype == FILETYPE_UNKNOWN) {
    errmsg("%s does not appear to be a G16 or TIFF file.", infile_name);
    return 8;
  }

  if (!outfile_name) {
#ifndef WIN32_GUI
    usage(); return 1;
#else
    char *ext;
    char *outfile_ext;
    DWORD result;

    memset(&ofn, sizeof(ofn), 0);
    outfile_name = malloc(PATH_MAX);

    if (filetype == FILETYPE_BMP) {
      outfile_ext = ".tiff";
      ofn.lpstrFilter = "TIFF Files (*.tiff)\0*.tiff;*.tif\0All Files (*.*)\0*.*\0";
    } else {
      outfile_ext = ".bmp";
      ofn.lpstrFilter = "G16 Files (*.bmp)\0*.bmp\0All Files (*.*)\0*.*\0";
    }

    ofn.lStructSize = sizeof(ofn);
    ofn.lpstrFile = outfile_name;
    ofn.nMaxFile = PATH_MAX;
    ofn.Flags = OFN_OVERWRITEPROMPT;
    ofn.lpstrDefExt = outfile_ext+1;

    strncpy(outfile_name, infile_name, PATH_MAX-5);
    ext = strrchr(outfile_name, '.');
    if (!ext ||
        (strcasecmp(ext, ".bmp") &&
         strcasecmp(ext, ".g16") &&
         strcasecmp(ext, ".tiff") &&
         strcasecmp(ext, ".tif"))) {
      ext = outfile_name + strlen(outfile_name);
    }
    strcpy(ext, outfile_ext);

    result = GetSaveFileName(&ofn);
    if (!result) {
      result = CommDlgExtendedError();
      if (result) {
        errmsg("Save dialog error: 0x%lx", result);
        return 9;
      } else {
        return 0;
      }
    }
#endif
  }

  if (filetype == FILETYPE_BMP) {
    infomsg("%s appears to be a G16 (BMP) file.  Converting to TIFF.", infile_name);
    if (read_bmp_headers(infile, &imginfo)) {
      return 3;
    }
    outfile = fopen(outfile_name, "wb");
    if (!outfile) {
      oserrmsg("Error opening output file");
      return 4;
    }
    if (write_tiff_headers(outfile, &imginfo)) {
      return 5;
    }
    for (row=0; row<imginfo.height; row++) {
      /* BMP is stored bottom to top, TIFF default (which we use) is top to
         bottom, so we need to reverse the order of the row data.. */
      if (read_g16_row(infile, &imginfo, imginfo.height - row - 1)) {
        return 6;
      }
      if (write_tiff_row(outfile, &imginfo)) {
        return 7;
      }
    }
    usermsg("Conversion to TIFF complete.");
  } else if (filetype == FILETYPE_TIFF) {
    infomsg("%s appears to be a TIFF file.  Converting to G16 (BMP).", infile_name);
    if (imginfo.convert_depth) {
      errmsg("16-bit to 8-bit conversion is only supported when converting from G16 format.");
      return 1;
    }
    if (read_tiff_headers(infile, &imginfo)) {
      return 3;
    }
    outfile = fopen(outfile_name, "wb");
    if (!outfile) {
      oserrmsg("Error opening output file");
      return 4;
    }
    if (write_bmp_headers(outfile, &imginfo)) {
      return 5;
    }
    for (row=0; row<imginfo.height; row++) {
      /* BMP is stored bottom to top.  If the TIFF file's orientation is
	 top-left (top to bottom, which is the default), we need to reverse the
	 order of the row data.  If the TIFF file is bottom-left (bottom to
         top) then it's the same as BMP, so we can just copy straight over.
	 (we don't support non-"left" orientations for TIFF.. too unlikely to
         be worth the effort) */
      if (read_tiff_row(infile, &imginfo, imginfo.bottom_up ? row : imginfo.height - row - 1)) {
        return 6;
      }
      if (write_g16_row(outfile, &imginfo)) {
        return 7;
      }
    }
    usermsg("Conversion to G16 complete.");
  }

  return 0;
}
