/* tgafix - Fix broken alpha info on TGA files exported from UnrealEd
 * Author: Alex Stewart <alex@foogod.com>
 * Date: Dec 3, 2004
 * 
 * This code has been placed into the public domain by the author.
 *
 * Description:
 *
 * UnrealEd (the map editor for the series of games based on the Unreal Engine)
 * can export textures in several different formats.  Unfortunately, the only
 * format to which it can export which has the capability of including
 * alpha-channel information is the TGA (Targa) format, and even more
 * unfortunately, UnrealEd writes broken TGA files so that the resulting files
 * claim not to have an alpha channel even though they do.
 *
 * Some graphics programs (wrongly) ignore the TGA file header and blithely
 * assume that there's an alpha channel and load the alpha data anyway, but
 * correctly written programs will load UnrealEd's TGA files without any alpha
 * information.  Thus, the one format which can hold alpha information still
 * doesn't work properly for exporting textures with alpha channels.
 *
 * This program fixes the TGA files produced by UnrealEd to correctly report
 * their alpha channel, so standards-conformant graphics software can load them
 * properly.
 *
 * To use this program, simply run it with the name of the TGA file you wish to
 * convert, and it will fix it.
 */

/* This program is fairly simplistic.  It only does enough to get the job done
   (well, and a little bit to make sure it's not screwing things up completely).
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define WINDOW_TITLE "UnrealEd TGA Alpha Fix"

#define TGA_POS_TYPE 2
#define TGA_POS_BPP  16
#define TGA_POS_ID   17
#define TGA_ID_MASK  0x0F
#define TGA_ID_FIX   0x08

#ifndef WIN32_GUI

/* We're compiling the normal command-line version.  Have messages go to
   stdout/stderr */

#define infomsg(fmt, args...) { printf(fmt "\n", ## args); fflush(stdout); }
#define usermsg(fmt, args...) { printf(fmt "\n", ## args); fflush(stdout); }
#define errmsg(fmt, args...) fprintf(stderr, "Error: " fmt "\n", ## args)
#define oserrmsg(str) perror(str)

#else

/* We're compiling this for the "_w" (Windows GUI mode) executable.  Do user
   interaction via dialog boxes and such */

#include <windows.h>

#define infomsg(fmt, args...) /* Don't show info messages in GUI mode */
#define usermsg(fmt, args...) dialog_box(MB_OK | MB_ICONINFORMATION, fmt, ## args)
#define errmsg(fmt, args...) dialog_box(MB_OK | MB_ICONERROR, fmt, ## args)
#define oserrmsg(str) dialog_box(MB_OK | MB_ICONERROR, "%s: %s", str, strerror(errno))

const char window_title[] = WINDOW_TITLE;

#define MSGBOX_BUFSIZE 1024

int dialog_box(int flags, const char *fmt, ...) {
  va_list ap;
  char buf[MSGBOX_BUFSIZE];

  va_start(ap, fmt);
  vsnprintf(buf, MSGBOX_BUFSIZE, fmt, ap);
  return MessageBox(NULL, buf, window_title, flags);
}
  
#endif

char *progname;

void usage(void) {
  errmsg("Usage: %s [--value n] filename.tga", progname);
}

int read_byte(FILE *f, int pos) {
  int c;

  if (fseek(f, pos, SEEK_SET)) {
    oserrmsg("Can't seek to position in file");
    return EOF;
  }
  c = fgetc(f);
  if (c == EOF) {
    oserrmsg("Error reading from file");
  }
  return c;
}

int write_byte(FILE *f, int pos, int c) {
  if (fseek(f, pos, SEEK_SET)) {
    oserrmsg("Can't seek to position in file");
    return EOF;
  }
  if (fputc(c, f) != c) {
    oserrmsg("Error writing to file");
    return EOF;
  }
  return 0;
}

int main(int argc, char *argv[]) {
  char *filename = NULL;
  FILE *f;
  int new_value = TGA_ID_FIX;
  int i, typecode, bpp, id, newid;

  progname = argv[0];

  for (i=1; i<argc; i++) {
    if (!strcmp(argv[i], "--value")) {
      if (++i >= argc) {
        usage();
        return 1;
      }
      new_value = atoi(argv[i]);
    } else if (filename) {
      usage();
      return 1;
    } else {
      filename = argv[i];
    }
  }

  if (!filename) {
    usage();
    return 1;
  }

  infomsg("Processing: %s", filename);

  f = fopen(filename, "r+b");
  if (!f) {
    oserrmsg("Can't open file");
    return 1;
  }

  /* Do some sanity checking.. */
  typecode = read_byte(f, TGA_POS_TYPE);
  bpp = read_byte(f, TGA_POS_BPP);
  if ((typecode != 2 && typecode != 10) || (bpp != 32)) {
    errmsg("%s does not appear to be a 32-bit (RGBA) TGA file.", filename);
    return 1;
  }
  
  id = read_byte(f, TGA_POS_ID);
  if (id == EOF) return 1;

  newid = (id & ~TGA_ID_MASK) | (new_value & TGA_ID_MASK);

  infomsg("Changing TGA image descriptor byte: old=%02x new=%02x", id, newid);
  
  if (write_byte(f, TGA_POS_ID, newid)) return 1;

  fclose(f);

  usermsg("%s alpha channel fixed", filename);
  return 0;
}
